' QBASIC/QB64 subset interpreter by Ed Davis.
' ------------------------------------------------------------------------------------------
Option _Explicit

$Console
$ScreenHide
_Dest _Console

Declare Library
    Function GetCommandLine Alias GetCommandLineA ()
End Declare
Dim Shared a As String
a$ = Command$(0)
Print a$
'------------------------------------------------------------------------
'  03 Jul 2021 todo
'  [x] store/retrieve variables like eval-ed4
'  [x] const id[$] = number|string {, const id[$] = number|string}
'  [ ] consolidate loop handling data structures
'  [x] arrays
'      [x]  parse
'      [x]  allocate 1 dimensional
'      [x]  allocate 2 dimensional
'      [x]  assign (idstmt, stridstmt)
'      [x]  reference (strfactor, primary)
'  [ ] Subs
'  [ ] Functions
'  [ ] Shared variables
'  getvarindex& (getstrindex$), used in:
'  forstmt:    to reference the "i" variable
'  inputstmt:  to reference the "input" variable: input "", numeric_store(i)
'  swapstmt:   reference: swap(numeric_store(i1), numeric_store(i2))
'  assignment: numeric_store(i) = value
'  primary:    primary# = numeric_store(i)
'idstmt (stridstmt) - only called by assignment
'------------------------------------------------------------------------
' Currently supports:
'
' Double and string variables only. No dim for these.
' One or two dimensional arrays.
' All standard operators, with hopefully the correct precedence.
'
' String operators:
' +
' =, <>, <, >, <=, >=
'
'
' Numeric binary operators
' ^,
' *, /
' \
' mod,
' +, -,
' shl,<<, shr,>>
' =, <>, <, >, <=, >=,
' and,
' or,
' xor,
' eqv,
' imp,
'
' Numeric unary operators
' not, -, +
'
' Commands supported:
'
' bye, quit - exits the interpreter
' clear     - clears variables
' dump      - shows info about arrays
' edit      - edits the current or default program
' files     - shows a list of files
' help      - simple help screen
' list      - lists current program
' list vars - iists variables
' load, old - loads a program into the interpreter
' new       - discards the current program
' reload    - reloads the current program from disk
' run       - runs the current program
' save      - saves the current program
' ston      - turns stepping on
' stoff     - turns stepping off
' tron      - turns tracing on
' troff     - turns tracing off
'
' Statements:
'
' Iteration:
'   do [while|until]
'     stmts
'   loop
'
'   or
'
'   do
'     stmts
'   loop [while|until]
'
'   exit do
'
'   while stmts wend - exit while
'
'   for index = n to expr [step n]
'   next
'
'   exit for
'
' Control transfer:
'   gosub, return
'   goto - buggy if you jump outside of loops or multiline if
'
' Selection:
' singleline or multiline if
'   if elseif else end if
'
' const - declare numeric or string constants
'
' dim - declares double or string arrays
'
' end, stop, system - exits program to interpreter
'
' Other commands:
'
' chdir circle cls color draw environ input line locate mid$ paint
' palette play preset print, ? pset randomize rem screen shell
' sleep sound swap view width window
'
' QB64 commands:
'
' _delay _display _freeimage _fullscreen _limit _printstring
' _screenmove _title
'
' Numeric functions:
'
' abs acos acosh acot acoth acsc acsch asc asec asech asin asinh
' atanh atn, atan cdbl cint clng cos cosh cot coth csc csch csng
' csrlin cvd cvi exp false frac fix instr int len ln log log10 peek
' point pos rnd screen sec sech sgn sin sinh sqr, sqrt tan tanh
' timer true ubound val
'
' QB64 Numeric functions:
'
' _atan2 _ceil _d2g _d2r _fontwidth _fontheight _g2d _g2r _height
' _instrrev _keydown _keyhit _mousebutton _mouseinput _mousex
' _mousey _newimage _pi _r2d _r2g _rgb _rgba _rgba32 _rgb32 _round
' _width
'
' String Functions:
'
' chr$ command$ date$ environ$ hex$ inkey$ lcase$ left$" lpad$
' ltrim$ mid$ mki$ oct$ replace$ right$ rpad$ rtrim$ space$ str$
' string$ time$ trim$ ucase$
'
' QB64 string functions:
'
' _clipboard$ _cwd$ _os$ _startdir$ _title$ _trim$
'
' No other numeric data types besides double. Only suffix accepted is $ for strings.
' Only up to two dimensional arrays
' No subs or functions.
' No dim.
' Lots of other stuff missing.
'------------------------------------------------------------------------------------------

Declare function accept&(s as string)
Declare function any_expr&(p as integer)
Declare function bool_expr&
Declare function numeric_expr#
Declare function numeric_expr2#(p as integer)
Declare function find_matching_else$
Declare function gettoeol$
Declare function instrfun&
Declare function is_multi_line_if&
Declare function is_stmt_end&
Declare function peek_ch$
Declare function pop_num#
Declare function pop_str$
Declare function primary#
Declare function storeline&
Declare function strexpression$
Declare function strfactor$

Declare sub clearprog
Declare sub clearvars
Declare sub colorstmt
Declare sub docmd(interactive as integer)
Declare sub elsestmt
Declare sub endifstmt
Declare sub exitstmt
Declare sub expect(s as string)
Declare sub filesstmt
Declare sub find_matching_sline_if
Declare sub find_matching_pair(s1 as string, s2 as string)
Declare sub forstmt
Declare sub getsym
Declare sub gosubline(target as integer)
Declare sub gosubstmt
Declare sub gotoline(target as integer)
Declare sub gotostmt
Declare sub idstmt
Declare sub ifstmt
Declare sub initvars
Declare sub inputstmt
Declare sub lineinputstmt
Declare sub linestmt
Declare sub liststmt
Declare sub loadprog(fn as string)
Declare sub locatestmt
Declare sub midstmt
Declare sub multi_ifstmt(cond as integer)
Declare sub nextstmt
Declare sub printstmt
Declare sub randomizer
Declare sub returnstmt
Declare sub saveprog
Declare sub showhelp
Declare sub skiptoeol
Declare sub stridstmt
Declare sub validlinenum(n as integer)
Declare sub wendstmt
Declare sub whilestmt(first as integer)

Const true = -1, false = 0
Const e = 2.71828182845905
Const halfpi = 1.5707963267949
Const pi = 3.14159265358979
Const max_store = 512
Const varssize = 64
Const stacksize = 512
Const pgmsize = 3000
Const tyunknown = 0, tyident = 1, tystrident = 2, tynum = 3, tystring = 4
Const left_side = 0, right_side = 1

Dim Shared the_ch As String ' last char read from input
Dim Shared sym As String ' last symbol read
Dim Shared symtype As Integer ' type of last symbol read
Dim Shared the_num As Double ' last number read
Dim Shared pgm(pgmsize) As String ' lines of text stored here
Dim Shared curline As Integer ' number of current line in pgm
Dim Shared thelin As String ' text of current line
Dim Shared textp As Integer ' positionn in thelin
Dim Shared errors As Integer
Dim Shared curr_filename As String

Type do_loop_t
    lline As Integer
    loff As Integer
End Type

' do/while/for/if tracking
Dim Shared loopvars(varssize) As Integer, looplines(varssize) As Integer
Dim Shared loopmax(varssize) As Double, loopstep(varssize) As Double
Dim Shared loopoff(varssize) As Integer
Dim Shared stackp As Integer, loopp As Integer
Dim Shared gosubstack(stacksize) As Integer, gosuboffstack(stacksize) As Integer
Dim Shared while_sp As Integer, while_line(varssize) As Integer, while_off(varssize) As Integer
Dim Shared do_sp As Integer, do_loop(stacksize) As do_loop_t
Dim Shared if_sp As Integer, if_stack(stacksize) As Integer

' for arrays:  make sure the user specified index is between lo_bnd..hi_bnd inclusive
' then, computed index = v.index + user_index - v.lo_bnd
Type names_t
    vname As String
    symtype As Integer ' variable name
    index As Long ' index into string table; numeric table; or string/numeric array table
    is_const As Integer
    lo_bnd As Long
    hi_bnd As Long
    lo_bnd2 As Long ' only if 2 dimensional
    hi_bnd2 As Long ' only if 2 dimensional
    multi As Integer ' true if 2 dimensional
    a_len As Long ' non-zero if array
    a_width As Long ' used in computing index when 2 dimensional
End Type

' variable names
Dim Shared var_names(1 To max_store) As names_t, var_names_max As Integer

' string and numeric values
Dim Shared string_store(1 To max_store) As String, str_store_max As Integer
Dim Shared numeric_store(1 To max_store) As Double, num_store_max As Integer

' string and numeric arrays
ReDim Shared string_arr_store(0) As String: Dim Shared str_arr_stor_max As Long
ReDim Shared numeric_arr_store(0) As Double: Dim Shared num_arr_stor_max As Long

' used by expression parser
Dim Shared str_stack(varssize) As String
Dim Shared num_stack(varssize) As Double
Dim Shared str_st_ndx As Integer, num_st_ndx As Integer

Dim Shared endif_count As Integer
Dim Shared wend_count As Integer
Dim Shared next_count As Integer
Dim Shared loop_count As Integer

Dim Shared tracing As Integer
Dim Shared stepping As Integer
Const right_assoc = 0, left_assoc = 1, unaryminus_prec = 13, unaryplus_prec = 13, unarynot_prec = 6

'for performance timing
'dim shared scantime as double
'dim shared starttime as double
'dim shared nsyms as long

Dim Shared ctype_arr(255) As Integer
Const ct_unknown = 0, ct_alpha = 1, ct_digit = 2, ct_period = 3, ct_punc1 = 4
Const ct_dquote = 5, ct_squote = 6, ct_amp = 7, ct_lt = 8, ct_gt = 9

'---------------------------------------------------------------------------------------------------
' Listed here since I can not remember them:
' % = integer (16 bit)
' & = long    (32 bit)
' ! = single  (default)
' # = double
' $ = string
'---------------------------------------------------------------------------------------------------
' Maybe add:
' min(x, x1, x2...), max(...), ave(...), sum(...)
'#define floor(x) ((x*2.0-0.5)shr 1)
'#define ceil(x) (-((-x*2.0-0.5)shr 1))
'---------------------------------------------------------------------------------------------------

Call init_scanner
tracing = false
stepping = false

str_st_ndx = 0
num_st_ndx = 0

'------------------------------------------------------------------------
' main loop
'------------------------------------------------------------------------
Dim cmd$
Dim quit As Integer
quit = false

'starttime = timer
cmd$ = Command$
If Command$(1) = "-t" Then
    cmd$ = LTrim$(RTrim$(Mid$(Command$, 4)))
    If cmd$ <> "" Then quit = true
Else
    _ScreenShow
    _Dest 0
End If

If cmd$ <> "" Then
    pgm(0) = "run " + cmd$
    Call initgetsym(0, 1)
    Call docmd(false)
Else
    Call showhelp
End If

If quit Then Call showtime: If errors Then System 1 Else System

Do
    Line Input "> ", pgm(0)
    If pgm(0) <> "" Then
        Call initgetsym(0, 1)
        Call docmd(true)
    End If
Loop

' show timings
Sub showtime
    'dim total_time as double
    'total_time = timer - starttime

    'print "Total time: "; total_time; " Scan time: "; scantime; " Parse time: "; total_time - scantime; " Symbols: "; nsyms
    'sleep
End Sub

Function at_line$
    at_line$ = "": If curline > 0 Then at_line$ = "(" + Str$(curline) + ")"
End Function

Function rest_of_line$
    rest_of_line$ = sym + " " + the_ch + " " + Mid$(thelin, textp)
End Function

Sub dump_tables
    Dim i As Integer

    Print "name", "index", "lo", "hi", "len"
    For i = 1 To var_names_max
        Print var_names(i).vname, var_names(i).index, var_names(i).lo_bnd, var_names(i).hi_bnd, var_names(i).a_len
    Next
End Sub

'------------------------------------------------------------------------
' command processor
'------------------------------------------------------------------------
Sub docmd (interactive As Integer)
    'print "docmd"
    errors = false

    restart_loop:
    stackp = 0 ' these were -1 ??? @review
    loopp = 0 ' these were -1 ??? @review
    while_sp = 0
    do_sp = 0
    if_sp = 0
    _AutoDisplay

    Do
        loop_top:
        If errors Then Exit Sub
        While sym = "" And curline > 0 And curline < pgmsize
            Call initgetsym(curline + 1, 1)
        Wend
        If tracing Then Print "["; curline; "] "; sym; " "; the_ch; " "; LTrim$(Mid$(thelin, textp))
        If stepping Then Sleep
        Select Case sym
            Case "": Exit Sub
            Case "bye", "quit": Call getsym: Call showtime: If errors Then System 1 Else System
            Case "clear": Call getsym: Call clearvars: Exit Sub
            Case "edit": Call getsym: Call editstmt: Exit Sub
            Case "help": Call getsym: Call showhelp: Exit Sub
            Case "list": Call getsym: Call liststmt: Exit Sub
            Case "load", "old": Call getsym: Call loadprog(""): Exit Sub
            Case "new": Call getsym: Call initvars: Call clearprog: tracing = false: Exit Sub
            Case "reload": Call getsym: Call loadprog(curr_filename): Exit Sub
            Case "run": Call getsym: Call runprog: interactive = false: GoTo restart_loop
            Case "save": Call getsym: Call saveprog: Exit Sub
            Case "stop", "end", "system": Call getsym: Exit Sub
            Case "dump": Call getsym: Call dump_tables

            Case "chdir": Call getsym: Call chdircmd
            Case "circle": Call getsym: Call circlestmt
            Case "cls": Call getsym: Cls
            Case "color": Call getsym: Call colorstmt
            Case "const": Call getsym: Call conststmt
            Case "dim": Call getsym: Call dimstmt
            Case "do": Call getsym: Call dostmt(true)
            Case "draw": Call getsym: Call drawstmt
            Case "else": Call getsym: Call elsestmt
            Case "elseif": Call getsym: Call elseifstmt
            Case "endif": Call getsym: Call endifstmt
            Case "environ": Call getsym: Call environstmt
            Case "exit": Call getsym: Call exitstmt
            Case "for": Call getsym: Call forstmt
            Case "gosub": Call getsym: Call gosubstmt: GoTo loop_top
            Case "goto": Call getsym: Call gotostmt: GoTo loop_top
            Case "if": Call getsym: Call ifstmt: GoTo loop_top
            Case "input": Call getsym: Call inputstmt
            Case "line": Call getsym: Call lineinputstmt
            Case "loop": Call getsym: Call loopstmt
            Case "locate": Call getsym: Call locatestmt
            Case "mid$": Call getsym: Call midstmt
            Case "next": Call getsym: Call nextstmt
            Case "paint": Call getsym: Call paintstmt
            Case "palette": Call getsym: Call palettestmt
            Case "preset": Call getsym: Call presetstmt
            Case "print", "?": Call getsym: Call printstmt
            Case "pset": Call getsym: Call psetstmt
            Case "randomize": Call getsym: Call randomizer
            Case "rem", "'": Call getsym: Call skiptoeol
            Case "return": Call getsym: Call returnstmt: GoTo loop_top
            Case "screen": Call getsym: Call screenstmt
            Case "shell": Call shellstmt
            Case "sleep": Call getsym: Call sleepstmt
            Case "swap": Call getsym: Call swapstmt
            Case "view": Call getsym: Call viewstmt
            Case "wend": Call getsym: Call wendstmt
            Case "while": Call getsym: Call whilestmt(true)
            Case "width": Call getsym: Call widthstmt
            Case "window": Call getsym: Call windowstmt

            Case "files": Call getsym: Call filesstmt
            Case "play": Call getsym: Play strexpression$
            Case "sound": Call getsym: Call soundstmt
            Case "_delay": Call getsym: _Delay numeric_expr#
            Case "_display": Call getsym: _Display
            Case "_freeimage": Call getsym: Call freeimage
            Case "_fullscreen": Call getsym: _FullScreen
            Case "_limit": Call getsym: Call limitstmt
            Case "_printstring": Call getsym: Call printstringstmt
            Case "_screenmove": Call getsym: Call screenmovestmt
            Case "_title": Call getsym: Call titlestmt

            Case "troff": Call getsym: tracing = false
            Case "tron": Call getsym: tracing = true
            Case "stoff": Call getsym: stepping = false
            Case "ston": Call getsym: stepping = true

                ' need to account for:
                '  - assignment
                '    let ...
                '    [str]ident = expression
                '    [str]ident(expression [, expression]) = expression
                '  - labels
                '    ident:
                '  - non-assignment, including labels
                '
            Case Else
                If Left$(sym, 1) = "_" Then
                    Print "Unknown command: "; sym: errors = true
                ElseIf accept&("let") Then
                    Call assignment
                ElseIf symtype = tyident Or symtype = tystrident Then
                    If peek_ch$ = "=" Then
                        Call assignment
                    ElseIf peek_ch$ = "(" And InStr(Mid$(pgm(curline), textp), "=") > 0 Then
                        Call array_assignment
                    ElseIf symtype = tyident And the_ch = ":" Then
                        Call getsym
                    ElseIf interactive Then
                        Call printstmt
                    ElseIf sym <> ":" And sym <> "" Then
                        Print at_line$; "Stmt expected, found:"; rest_of_line$: errors = true
                    End If
                ElseIf interactive Then
                    Call printstmt
                ElseIf sym <> ":" And sym <> "" Then
                    Print at_line$; "Stmt expected, found:"; rest_of_line$: errors = true
                End If
        End Select

        If errors Then
            Exit Sub
        ElseIf sym = ":" Then
            Call getsym
        ElseIf sym <> "else" And sym <> "" Then
            Print at_line$; "Extra stmts:"; rest_of_line$: errors = true
            Print "symtype:"; symtype; " sym:"; sym; " ch:"; the_ch
        End If
    Loop
End Sub

'------------------------------------------------------------------------
' variable storage/retrieval
'------------------------------------------------------------------------

' find position of vname in var_names
Function find_vname& (vname As String)
    Dim i As Integer

    For i = 1 To var_names_max
        If var_names(i).vname = vname Then
            find_vname& = i
            Exit Function
        End If
    Next

    find_vname& = 0
End Function

' helper function for 2d arrays
Function ar_scale (i As Long, lo As Long)
    ar_scale = i - (lo - 1)
End Function

' get the index of "a" in either string_arr_store Or numeric_arr_store
' pointing to: a(expr [, expr])
Function get_array_index& (ident As String)
    Dim i As Integer, index As Long, index2 As Long, lo As Long, lo2 As Long
    Dim x As Long

    Call getsym
    i = find_vname&(ident)
    If i = 0 Then Print at_line$; "Array has not been declared: "; ident: errors = true: Exit Function
    expect ("(")
    index = numeric_expr#

    If var_names(i).multi Then
        expect (",")
        index2 = numeric_expr#
    End If

    expect (")")
    If var_names(i).a_len = 0 Then
        Print at_line$; "'"; ident; "' is not declared as an array": errors = true: Exit Function
    End If

    ' verfiy that the index is within range
    If index < var_names(i).lo_bnd Or index > var_names(i).hi_bnd Then
        Print at_line$; "Index is out of range:"; index; "("; var_names(i).lo_bnd; ","; var_names(i).hi_bnd; ")": errors = true: Exit Function
    End If
    If var_names(i).multi Then
        If index2 < var_names(i).lo_bnd2 Or index2 > var_names(i).hi_bnd2 Then
            Print at_line$; "Index two is out of range:"; index2: errors = true: Exit Function
        End If
    End If

    ' compute the actual index
    lo = var_names(i).lo_bnd
    If var_names(i).multi Then
        lo2 = var_names(i).lo_bnd2
        x = var_names(i).index + (var_names(i).a_width * (ar_scale(index2, lo2) - 1) + ar_scale(index, lo)) - 1
    Else
        x = var_names(i).index + ar_scale(index, lo) - 1
        'x = var_names(i).index + index - (var_names(i).lo_bnd - 1) - 1
    End If
    'print "index: "; x

    If tracing Then Print ident; "("; index; ")["; x; "] =";
    get_array_index& = x
End Function

' primary: if var does not exist, create it.  Return the var store index
' sym is the numeric variable name
Function getvarindex& (side As Integer)
    Dim i As Integer, ident As String, ident_type As Integer

    ident = sym: ident_type = symtype
    Call getsym

    If ident_type = tystrident Then
        Print at_line$; "type mismatch": errors = true
    ElseIf ident_type <> tyident Then
        Print at_line$; "not a variable": errors = true
    Else
        ' see if variable exists
        i = find_vname&(ident)
        If i > 0 Then
            getvarindex& = var_names(i).index
            If side = left_side And var_names(i).is_const Then Print at_line$; "Cannot update const variable: "; ident: errors = true
            Exit Function
        End If

        'if side = right_side then print at_line$; "Reference to unassigned variable: "; ident: errors = true

        ' create a new variable
        num_store_max = num_store_max + 1
        var_names_max = var_names_max + 1

        var_names(var_names_max).vname = ident
        var_names(var_names_max).symtype = ident_type
        var_names(var_names_max).index = num_store_max
        numeric_store(num_store_max) = 0 ' default value

        getvarindex& = num_store_max
    End If
End Function

Function getstrindex& (side As Integer)
    Dim i As Integer, ident As String, ident_type As Integer

    ident = sym: ident_type = symtype
    Call getsym

    If ident_type = tyident Then
        Print at_line$; "type mismatch": errors = true
    ElseIf ident_type <> tystrident Then
        Print at_line$; "not a variable": errors = true
    Else
        ' see if variable exists
        i = find_vname&(ident)
        If i > 0 Then
            getstrindex& = var_names(i).index
            If side = left_side And var_names(i).is_const Then Print at_line$; "Cannot update const variable: "; ident: errors = true
            Exit Function
        End If

        'if side = right_side then print at_line$; "Reference to unassigned variable: "; ident: errors = true

        ' create a new variable
        str_store_max = str_store_max + 1
        var_names_max = var_names_max + 1

        var_names(var_names_max).vname = ident
        var_names(var_names_max).symtype = ident_type
        var_names(var_names_max).index = str_store_max
        string_store(str_store_max) = "" ' default value

        getstrindex& = str_store_max
    End If
End Function

' a(expr)
' when called, sym pointing at the ident
Function get_numeric_array_value#
    Dim ident As String, ident_type As Integer, x As Long, n As Double

    ident = sym: ident_type = symtype
    x = get_array_index&(ident)
    n = numeric_arr_store(x)
    get_numeric_array_value# = n
End Function

' a(expr)
' when called, sym pointing at the ident
Function get_string_array_value$
    Dim ident As String, ident_type As Integer, x As Long, s As String

    ident = sym: ident_type = symtype
    x = get_array_index&(ident)
    s = string_arr_store(x)
    get_string_array_value$ = s
End Function

Sub stridstmt
    Dim i As Integer, vname As String

    vname = sym
    'print "stridstmt"
    i = getstrindex&(left_side)
    expect ("=")
    string_store(i) = strexpression$
    If tracing Then Print vname, string_store(i)
End Sub

Sub idstmt
    Dim i As Integer, vname As String

    vname = sym
    i = getvarindex&(left_side)
    expect ("=")
    numeric_store(i) = numeric_expr#
    If tracing Then Print vname, numeric_store(i)
End Sub

' ident = expression
Sub assignment
    If symtype = tyident Then
        Call idstmt
    ElseIf symtype = tystrident Then
        Call stridstmt
    Else
        Print at_line$; "Expecting assignment stmt, found: "; sym: errors = true
    End If
End Sub

' ident(expression [, expression]) = expression
Sub array_assignment
    Dim ident As String, ident_type As Integer
    Dim s As String, n As Double, x As Long

    ident = sym: ident_type = symtype
    x = get_array_index&(ident)

    expect ("=")

    If ident_type = tystrident Then
        s = strexpression$
        'assign string
        string_arr_store(x) = s
        If tracing Then Print s
    Else
        n = numeric_expr#
        'assign number
        numeric_arr_store(x) = n
        If tracing Then Print n
    End If
End Sub

'------------------------------------------------------------------------
' statement parsing
'------------------------------------------------------------------------

Sub showhelp
    Print "bye or quit -- exit"
    Print "help        -- show this screen"
    Print "clear       -- clear variables"
    Print "edit        -- edit current program"
    Print "list        -- show source"
    Print "list vars   -- show variables"
    Print "load        -- load program from disk"
    Print "new         -- clear program in memory"
    Print "reload      -- reload program from disk"
    Print "run         -- run program in memory"
    Print "save        -- save program to disk"
    Print ""
    Print "cls         -- clear screen"
    Print "tron        -- tracing on"
    Print "troff       -- tracing off"
    Print "ston        -- stepping on"
    Print "stoff       -- stepping off"
End Sub

Function getfn$ (prompt As String)
    Dim filespec As String
    If symtype = tystring Or symtype = tystrident Then
        filespec = strexpression$
    ElseIf sym <> "" Then
        filespec = sym ' gettoeol destroys sym
        filespec = filespec + gettoeol$
    Else
        Print prompt; ": ";
        Line Input filespec
    End If
    If filespec <> "" Then
        If InStr(filespec, ".") = 0 Then filespec = filespec + ".bas"
    End If
    getfn$ = filespec
End Function

Sub clearvars
    Dim i As Integer

    For i = 1 To str_store_max
        string_store(i) = ""
    Next

    For i = 1 To num_store_max
        numeric_store(i) = 0
    Next
End Sub

Sub initvars
    Dim i As Integer

    clearvars
    For i = 1 To var_names_max
        var_names(i).vname = ""
        var_names(i).index = 0
    Next

    str_store_max = 0: num_store_max = 0: var_names_max = 0
End Sub

Sub loadprog (fn As String)
    Dim n As Integer
    initvars
    clearprog

    If fn = "" Then curr_filename = getfn$("Program file") Else curr_filename = fn
    If curr_filename = "" Then Exit Sub
    Open curr_filename For Input As 1

    n = 0
    While Not EOF(1)
        Line Input #1, pgm(0)
        'if pgm(0) <> "" then
        If storeline& Then
            n = the_num + 1
        Else
            n = n + 1
            pgm(n) = pgm(0)
        End If
        'end if
    Wend

    Close #1
    curline = 0
End Sub

Sub editstmt
    Dim editor As String
    editor = Environ$("EDITOR")
    If editor = "" Then editor = "notepad.exe"
    If curr_filename = "" Then curr_filename = "default.bas"
    Shell editor + " " + curr_filename
    Call loadprog(curr_filename)
End Sub

Sub runprog
    If sym <> "" Then Call loadprog("")
    Call initgetsym(1, 1)
End Sub

Sub saveprog
    Dim filespec As String
    Dim i As Integer
    filespec = getfn$("Save as")
    If filespec = "" Then Exit Sub
    Open filespec For Output As 1
    If Err = 8 Then
        Print at_line$; "*** error: you don't have permission to write to that file."
        Exit Sub
    End If
    For i = 1 To pgmsize
        If Len(pgm(i)) Then Print #1, i; " "; pgm(i)
    Next
    Close #1
End Sub

Sub liststmt
    Dim i As Integer
    If sym = "vars" Then
        For i = 1 To var_names_max
            If var_names(i).a_len > 0 Then
                Print "Array:"; var_names(i).vname, " index: "; var_names(i).index;
                Print string_store(var_names(i).index); " size: "; var_names(i).a_len;
                Print " type: "; var_names(i).symtype
            ElseIf Right$(var_names(i).vname, 1) = "$" Then
                Print "String:"; var_names(i).vname, " index: "; var_names(i).index;
                Print string_store(var_names(i).index);
                Print " type: "; var_names(i).symtype
            ElseIf var_names(i).vname <> "" Then
                Print "Number:"; var_names(i).vname, " index: "; var_names(i).index;
                Print numeric_store(var_names(i).index);
                Print " type: "; var_names(i).symtype
            End If
        Next
    Else
        For i = 1 To pgmsize
            If pgm(i) <> "" Then Print i; " "; pgm(i)
        Next
    End If
    Print
End Sub

Sub chdircmd
    ChDir strexpression$
End Sub

' CIRCLE [STEP] (x!,y!),radius![,[color%] [,[start!] [,[end!] [,aspect!]]]]
Sub circlestmt
    Dim x As Single, y As Single, radius As Single, clr As Long
    Dim arcbeg As Single, arcend As Single, elipse As Single

    expect ("(")
    x = numeric_expr#
    expect (",")
    y = numeric_expr#
    expect (")")
    expect (",")
    radius = numeric_expr#

    '[,[color%] [,[start!] [,[end!] [,aspect!]]]]
    If accept&(",") Then ' color
        If accept&(",") Then ' arcbeg
            If accept&(",") Then ' arcend
                If accept&(",") Then ' elipse
                    elipse = numeric_expr#
                    Circle (x, y), radius, , , , elipse
                Else
                    arcend = numeric_expr#
                    If accept&(",") Then
                        elipse = numeric_expr#
                        Circle (x, y), radius, , , arcend, elipse
                    Else
                        Circle (x, y), radius, , , arcend
                    End If
                End If
            Else
                arcbeg = numeric_expr#
                If accept&(",") Then ' arcend
                    If accept&(",") Then ' elipse
                        elipse = numeric_expr#
                        Circle (x, y), radius, , arcbeg, , elipse
                    Else
                        arcend = numeric_expr#
                        If accept&(",") Then
                            elipse = numeric_expr#
                            Circle (x, y), radius, , arcbeg, arcend, elipse
                        Else
                            Circle (x, y), radius, , arcbeg, arcend
                        End If
                    End If
                End If
            End If
        Else
            ' [,[start!] [,[end!] [,aspect!]]]]
            clr = numeric_expr#
            If accept&(",") Then ' arcbeg
                If accept&(",") Then ' arcend
                    If accept&(",") Then ' elipse
                        elipse = numeric_expr#
                        Circle (x, y), radius, clr, , , elipse
                    Else
                        arcend = numeric_expr#
                        If accept&(",") Then
                            elipse = numeric_expr#
                            Circle (x, y), radius, clr, , arcend, elipse
                        Else
                            Circle (x, y), radius, clr, , arcend
                        End If
                    End If
                Else
                    arcbeg = numeric_expr#
                    If accept&(",") Then ' arcend
                        If accept&(",") Then ' elipse
                            elipse = numeric_expr#
                            Circle (x, y), radius, clr, arcbeg, , elipse
                        Else
                            arcend = numeric_expr#
                            If accept&(",") Then
                                elipse = numeric_expr#
                                Circle (x, y), radius, clr, arcbeg, arcend, elipse
                            Else
                                Circle (x, y), radius, clr, arcbeg, arcend
                            End If
                        End If
                    Else
                        Circle (x, y), radius, clr, arcbeg
                    End If
                End If
            Else
                Circle (x, y), radius, clr
            End If
        End If
    Else
        Circle (x, y), radius
    End If
End Sub

' color [fore] [,back]
Sub colorstmt
    Dim back As Long, fore As Long
    If accept&(",") Then
        back = numeric_expr#
        Color , back
    Else
        fore = numeric_expr#
        If accept&(",") Then
            back = numeric_expr#
            Color fore, back
        Else
            Color fore
        End If
    End If
End Sub

Sub get_array_bounds (lo As Long, hi As Long)
    lo = numeric_expr#
    If accept&("to") Then
        hi = numeric_expr#
    Else
        hi = lo
        lo = 0
    End If
End Sub

' dim ident(numeric expression [to numeric expression]) {, ident(numeric expression [to numeric expression])}
Sub dimstmt
    Dim ident As String, ident_type As Integer, lo As Long, hi As Long, lo2 As Long, hi2 As Long
    Dim a_len As Long, index As Long, i As Integer, multi As Integer, a_width As Long

    Do
        ident = sym
        ident_type = symtype
        If symtype <> tyident And symtype <> tystrident Then
            Print at_line$; " Expecting an identifier, but found: "; sym: errors = true: Exit Sub
        End If
        Call getsym ' skip array name

        expect ("(")
        Call get_array_bounds(lo, hi)
        lo2 = 0: hi2 = 0: multi = false
        If accept&(",") Then Call get_array_bounds(lo2, hi2): multi = true
        expect (")")

        ' see if it already exists
        i = find_vname&(ident)
        If i > 0 Then Print "Duplicate definition: "; ident: errors = true: Exit Sub

        ' add it
        a_len = hi - lo + 1
        If multi Then
            a_width = a_len
            a_len = a_len * (hi2 - lo2 + 1)
        End If

        var_names_max = var_names_max + 1
        var_names(var_names_max).vname = ident
        var_names(var_names_max).symtype = ident_type
        var_names(var_names_max).lo_bnd = lo
        var_names(var_names_max).hi_bnd = hi
        var_names(var_names_max).lo_bnd2 = lo2
        var_names(var_names_max).hi_bnd2 = hi2
        var_names(var_names_max).multi = multi
        var_names(var_names_max).a_len = a_len
        var_names(var_names_max).a_width = a_width

        If ident_type = tystrident Then
            index = str_arr_stor_max + 1
            str_arr_stor_max = str_arr_stor_max + a_len
            ReDim _Preserve string_arr_store(str_arr_stor_max)
        Else
            index = num_arr_stor_max + 1
            num_arr_stor_max = num_arr_stor_max + a_len
            ReDim _Preserve numeric_arr_store(num_arr_stor_max)
        End If

        var_names(var_names_max).index = index

        If sym <> "," Then Exit Do
        Call getsym
    Loop
End Sub

' const id[$] = number|string {, const id[$] = number|string}
Sub conststmt
    Dim i As Integer

    Do
        i = find_vname&(sym)
        If i <> 0 Then Print at_line$; "var: "; sym; " already defined": errors = true: Exit Sub
        Call assignment
        var_names(var_names_max).is_const = true
    Loop While accept&(",")

End Sub

Sub drawstmt
    Dim s As String
    s = strexpression$
    Draw s
End Sub

Sub environstmt
    Environ strexpression$
End Sub

' need to account for loop [until|while expr] and next [i]
Sub exitstmt
    If sym = "while" Then
        Call getsym
        If while_sp <= 0 Then errors = true: Print at_line$; "'exit while' without while": errors = true: Exit Sub
        while_sp = while_sp - 1
        Call find_matching_pair("while", "wend")
        Call getsym
    ElseIf sym = "do" Then
        Call getsym
        If do_sp <= 0 Then errors = true: Print at_line$; "'exit do' without do": errors = true: Exit Sub
        do_sp = do_sp - 1
        Call find_matching_pair("do", "loop")
        Call getsym
        If sym = "until" Or sym = "while" Then
            Call getsym ' skip until\while
            ' somehow skip over the until\while expression
            While sym <> ":" And sym <> ""
                Call getsym
            Wend
        End If
    ElseIf sym = "for" Then
        Call getsym
        If loopp <= 0 Then errors = true: Print at_line$; "'exit for' without do": errors = true: Exit Sub
        loopp = loopp - 1
        Call find_matching_pair("for", "next")
        Call getsym
        If symtype = tyident Then Call getsym
    Else
        Print at_line$; "'exit without do/for/while": errors = true: Exit Sub
    End If

    If endif_count > 0 And if_sp > 0 Then if_sp = if_sp - endif_count
    If loop_count > 0 And do_sp > 0 Then do_sp = do_sp - loop_count
    If next_count > 0 And loopp > 0 Then loopp = loopp - next_count
    If wend_count > 0 And while_sp > 0 Then while_sp = while_sp - wend_count

End Sub

' for xvar = -1.5 to 1.5 step .01
Sub forstmt
    Dim xvar As Integer, i As Integer
    xvar = getvarindex&(left_side) ' get position of "i"
    If loopp >= 0 Then
        For i = 0 To loopp - 1
            If loopvars(i) = xvar Then
                Print at_line$; "for index variable already in use": errors = true
                Exit Sub
            End If
        Next
    End If
    expect ("=")
    numeric_store(xvar) = numeric_expr#
    loopp = loopp + 1
    loopvars(loopp) = xvar
    looplines(loopp) = curline
    expect ("to")
    loopmax(loopp) = numeric_expr#
    If accept&("step") Then loopstep(loopp) = numeric_expr# Else loopstep(loopp) = 1
    loopoff(loopp) = textp
    If Len(sym) > 0 Then loopoff(loopp) = textp - Len(sym) - 1
End Sub

' finds target, using current sym
Function get_target&
    If symtype = tynum Then
        get_target = numeric_expr#
    Else
        Dim i As Integer, lbl As String
        lbl = sym
        If Right$(lbl, 1) <> ":" Then lbl = lbl + ":"
        For i = 1 To pgmsize
            If LCase$(Mid$(LTrim$(pgm(i)), 1, Len(lbl))) = lbl Then
                get_target& = i
                Exit Function
            End If
        Next
        Print at_line$; "Target of goto not found:"; sym: errors = true
        get_target& = 0
    End If
End Function

Sub gosubstmt
    Dim target As Integer

    target = get_target&
    If Not errors Then
        validlinenum (target)
        stackp = stackp + 1
        If stackp > stacksize Then Print at_line$; "out of stack space": errors = true
        gosubstack(stackp) = curline
        ' 26 May 2021 was just textp
        gosuboffstack(stackp) = textp - 1
        'print "textp:"; textp; "=>"; pgm$(curline)
        'if sym = ":" then gosuboffstack(stackp) = textp
        Call initgetsym(target, 1)
    End If
End Sub

Sub gotostmt
    Dim target As Integer

    target = get_target&
    gotoline (target)
End Sub

' single line if: if expr then if expr then if expr then s else s else s else s
Sub ifstmt
    Dim b As Integer, level As Integer, cond As Integer

    level = 0

    begin:

    level = level + 1
    cond = numeric_expr#
    b = accept&("then")
    ' *** multiline if processing ***
    If sym = "" Then 'multiline if
        If level > 1 Then
            Print at_line$; "can't mix multi and single line 'if'": errors = true
            Exit Sub
        End If
        Call multi_ifstmt(cond)
        ' *** singleline if processing ***
    ElseIf cond Then
        If symtype = tynum Then
            gotoline (Int(the_num))
        ElseIf accept&("if") Then
            GoTo begin
        End If
    Else
        Call find_matching_sline_if
        ' if else found, pick up there, otherwise skip rest of stmt
        If Not accept&("else") Then skiptoeol
        If symtype = tynum Then gotoline (Int(the_num))
    End If
End Sub

Sub multi_ifstmt (cond As Integer)
    Dim s As String, b As Integer

    if_sp = if_sp + 1
    if_stack(if_sp) = curline
    'print at_line$; "if after inc: if_sp: "; if_sp, pgm(curline)

    If cond Then
        ' let docmd process these commands
    Else
        'need to find the next corresponding 'elseif' or 'else' or 'endif'
        restart:
        ' on the "if" Or "elseif" line, so skip it
        s = find_matching_else$ 'either elseif, else or endif
        'print at_line$; "found: "; sym
        If s = "" Then Print at_line$; "missing endif": errors = true: Exit Sub
        If tracing Then Print "["; curline; "] "; sym; " "; Mid$(thelin, textp)
        If sym = "elseif" Then
            'print sym; ": "; mid$(thelin, textp)
            Call getsym 'skip "elseif"
            cond = numeric_expr#
            b = accept&("then")
            'print at_line$; "elseif evaluated to: "; cond
            If cond Then
                ' let docmd process these commands, until next elseif/else/endif
            Else
                GoTo restart
            End If
        ElseIf sym = "else" Then
            'rem - let docmd process these commands, until endif
            Call getsym ' skip the else, so docmd goes to next line
        ElseIf sym = "endif" Then
            Call endifstmt
        End If
    End If
End Sub

' called from docmd()
Sub elseifstmt
    Dim s As String

    If if_sp = 0 Then Print at_line$; "endif without if": errors = true: Exit Sub

    'scan until matching endif

    ' but first, allow more "elseif"'s
    Do
        s = find_matching_else$
    Loop While s = "elseif"

    ' allow an "else"
    If s = "else" Then
        s = find_matching_else$
    End If

    ' finally, need an "endif"
    If s = "endif" Then
        ' pop the if stack
        if_sp = if_sp - 1
        Call getsym ' skip "endif"
        ' done
    Else
        Print at_line$; "Missing endif": errors = true: Exit Sub
    End If
End Sub

' called from docmd()
Sub elsestmt
    'print at_line$; "else begin: if_sp: "; if_sp, pgm(curline)

    'part of a single-line if?
    Call initgetsym(curline, 1)
    'if not "else", then single-line If
    If sym <> "else" Then Call skiptoeol: Exit Sub

    ' looks like multiline if - but have we seen the start of it?
    If if_sp = 0 Then Print at_line$; "else without if": errors = true: Exit Sub

    'scan until matching endif
    If find_matching_else$ <> "endif" Then Print at_line$; "else without endif": errors = true: Exit Sub

    Call getsym 'skip the "endif"
    'pop the if stack
    if_sp = if_sp - 1
    'print at_line$; "else end: if_sp: "; if_sp, pgm(curline)
End Sub

' called from docmd()
Sub endifstmt
    If if_sp = 0 Then Print at_line$; "endif without if": errors = true: Exit Sub
    if_sp = if_sp - 1
    'print at_line$; "endif: if_sp: "; if_sp, pgm(curline)
    Call getsym 'skip "endif"
End Sub

' input [;] ["prompt" ;|,] variablelist
Sub inputsetup
    If Left$(sym, 1) = Chr$(34) Then
        Print Mid$(sym, 2, Len(sym) - 1);
        Call getsym
        If accept&(";") Then
            Print "? ";
        Else
            expect (",")
        End If
    End If
End Sub

' input [;] ["prompt" ;|,] variablelist
Sub inputstmt
    Dim ident As String, ident_type As Integer, i As Long, x As Long, st As String, n As Double

    inputsetup

    ident = sym: ident_type = symtype
    If ident_type = tystrident Then
        Input "", st
    Else
        Input "", n
    End If

    i = find_vname&(ident)
    If i > 0 Then
        If ident_type <> var_names(i).symtype Then
            Print at_line$; "Type mismatch: "; ident_type; " vs. table: "; var_names(i).symtype: errors = true: Exit Sub
        End If
        If var_names(i).a_len > 0 Then ' array
            x = get_array_index&(ident)

            If ident_type = tystrident Then
                'assign string
                string_arr_store(x) = st
                If tracing Then Print st
            Else
                'assign number
                numeric_arr_store(x) = n
                If tracing Then Print n
            End If

            Exit Sub

        End If
    End If

    If ident_type = tystrident Then
        i = getstrindex&(left_side)
        string_store(i) = st
    ElseIf ident_type = tyident Then
        i = getvarindex&(left_side)
        numeric_store(i) = n
    Else
        Print at_line$; "Unknown type": errors = true: Exit Sub
    End If
End Sub

' line input [;] ["prompt";] variable$
Sub lineinputstmt
    Dim ident As String, ident_type As Integer, i As Long, x As Long, st As String

    If Not accept&("input") Then linestmt: Exit Sub
    inputsetup

    ident = sym: ident_type = symtype

    If ident_type <> tystrident Then Print at_line$; "String variable expected": errors = true: Exit Sub

    Line Input st

    i = find_vname&(ident)
    If i > 0 Then
        If ident_type <> var_names(i).symtype Then
            Print at_line$; "Type mismatch: "; ident_type; " vs. table: "; var_names(i).symtype: errors = true: Exit Sub
        End If
        If var_names(i).a_len > 0 Then ' array
            x = get_array_index&(ident)

            'assign string
            string_arr_store(x) = st
            If tracing Then Print st

            Exit Sub

        End If
    End If

    i = getstrindex&(left_side)
    string_store(i) = st
End Sub

' line [[step](x1!,y1!)]-[step](x2!,y2!) [,[color%] [,[b | bf] [,style%]]]
' ??? step is not currently supported
Sub linestmt
    Dim x1 As Single, y1 As Single, x2 As Single, y2 As Single, clr As Long
    Dim rect_type As String, step1 As Integer, step2 As Integer

    step1 = false: step2 = false

    If accept&("step") Then step1 = true
    expect ("(")
    x1 = numeric_expr#
    expect (",")
    y1 = numeric_expr#
    expect (")")

    expect ("-")
    If accept&("step") Then step2 = true

    expect ("(")
    x2 = numeric_expr#
    expect (",")
    y2 = numeric_expr#
    expect (")")

    ' so far we have: line(x, y)-(x2, y2)

    If is_stmt_end& Then Line (x1, y1)-(x2, y2): Exit Sub

    '[,[color%] [,[b | bf] [,style%]]]
    ' only acceptable value is a ","

    '1) ,c
    '2) ,c,b
    '3) ,c,b,s
    '4) ,c,,s
    '5) ,,b
    '6) ,,b,s
    '7) ,,,s

    If accept&(",") Then
        If accept&(",") Then
            If accept&(",") Then
                'must have s (7)
                Line (x1, y1)-(x2, y2), , , numeric_expr#
            Else
                'must have b
                rect_type = UCase$(sym)
                If rect_type <> "B" And rect_type <> "BF" Then
                    Print at_line$; "line ... - expecting 'B' or 'BF', found: "; rect_type: errors = true: Exit Sub
                End If
                If accept&(",") Then
                    'must have s (6)
                    If rect_type = "B" Then
                        Line (x1, y1)-(x2, y2), , B , numeric_expr#
                    Else
                        Line (x1, y1)-(x2, y2), , BF , numeric_expr#
                    End If
                Else
                    '(5)
                    If rect_type = "B" Then
                        Line (x1, y1)-(x2, y2), , B
                    Else
                        Line (x1, y1)-(x2, y2), , BF
                    End If
                    Call getsym ' skip "B"
                End If
            End If
        Else
            'must have c
            clr = numeric_expr#
            If accept&(",") Then
                If accept&(",") Then
                    'must have s (4)
                    Line (x1, y1)-(x2, y2), clr, , numeric_expr#
                Else
                    'must have b
                    rect_type = UCase$(sym)
                    If rect_type <> "B" And rect_type <> "BF" Then
                        Print at_line$; "line ... - expecting 'B' or 'BF', found: "; rect_type: errors = true: Exit Sub
                    End If
                    If accept&(",") Then
                        'must have s (3)
                        If rect_type = "B" Then
                            Line (x1, y1)-(x2, y2), clr, B , numeric_expr#
                        Else
                            Line (x1, y1)-(x2, y2), clr, BF , numeric_expr#
                        End If
                    Else
                        '(2)
                        If rect_type = "B" Then
                            Line (x1, y1)-(x2, y2), clr, B
                        Else
                            Line (x1, y1)-(x2, y2), clr, BF
                        End If
                        Call getsym ' skip "B"
                    End If
                End If
            Else
                '(1)
                Line (x1, y1)-(x2, y2), clr
            End If
        End If
    End If
End Sub

Sub locatestmt
    Dim row As Integer, col As Integer
    If accept&(",") Then
        col = numeric_expr#
        Locate , col
    Else
        row = numeric_expr#
        If accept&(",") Then
            col = numeric_expr#
            Locate row, col
        Else
            Locate row
        End If
    End If
End Sub

' mid$(s, i, n)
Sub midstmt
    Dim xvar As Integer, start As Integer, length As Integer, nolength As Integer
    expect ("(")
    xvar = getstrindex&(left_side)
    expect (",")
    start = numeric_expr#
    If accept&(",") Then length = numeric_expr# Else nolength = -1
    expect (")")
    expect ("=")
    If nolength Then
        Mid$(string_store(xvar), start) = strexpression$
    Else
        Mid$(string_store(xvar), start, length) = strexpression$
    End If
End Sub

Sub nextstmt
    Dim cont As Integer

    If symtype = tyident Then Call getsym
    If loopp < 0 Then Print at_line$; "next without for": errors = true: Exit Sub
    ' increment the current "i"
    numeric_store(loopvars(loopp)) = numeric_store(loopvars(loopp)) + loopstep(loopp)
    If tracing Then Print "["; curline; "] "; "next: "; numeric_store(loopvars(loopp))

    ' see if "for" should continue
    cont = false
    If loopstep(loopp) < 0 Then
        If numeric_store(loopvars(loopp)) >= loopmax(loopp) Then
            cont = true
        End If
    Else
        If numeric_store(loopvars(loopp)) <= loopmax(loopp) Then
            cont = true
        End If
    End If

    If cont Then
        Call initgetsym(looplines(loopp), loopoff(loopp))
    Else
        loopp = loopp - 1
    End If

End Sub

' PAINT [STEP] (column%, row%), fillColor[, borderColor%]
Sub paintstmt
    Dim x As Long, y As Long, f As Long

    expect ("(")
    x = numeric_expr#
    expect (",")
    y = numeric_expr#
    expect (")")
    If accept&(",") Then
        If accept&(",") Then
            Paint (x, y), , numeric_expr#
        Else
            f = numeric_expr#
            If accept&(",") Then
                Paint (x, y), f, numeric_expr#
            Else
                Paint (x, y), f
            End If
        End If
    Else
        Paint (x, y)
    End If
End Sub

' palette [attribute%,color&]
Sub palettestmt
    Dim a As Integer, c As Long

    a = numeric_expr#
    expect (",")
    c = numeric_expr#
    Palette a, c
End Sub

Sub printstmt
    Dim val_type As Integer, printed As Integer

    printed = false
    If accept&(",") Then Print ,
    Do While sym <> "" And sym <> ":" And sym <> "else"
        printed = true
        val_type = any_expr&(0)

        If accept&(",") Then
            If val_type = tystring Then
                Print pop_str$,
            Else
                Print pop_num#,
            End If
        ElseIf accept&(";") Then
            If val_type = tystring Then
                Print pop_str$;
            Else
                Print pop_num#;
            End If
        Else
            If val_type = tystring Then
                Print pop_str$
            Else
                Print pop_num#
            End If
            Exit Do
        End If
    Loop
    If Not printed Then Print
End Sub

' preset   (column, row)
' preset [step] (x!,y!) [,color%]
Sub presetstmt
    Dim x As Single, y As Single
    expect ("(")
    x = numeric_expr#
    expect (",")
    y = numeric_expr#
    expect (")")
    If accept&(",") Then PReset (x, y), numeric_expr# Else PReset (x, y)
End Sub

' pset   (column, row)
' pset [step] (x!,y!) [,color%]
' PSET [STEP] (x!,y!) [,color%]
Sub psetstmt
    Dim x As Single, y As Single, clr As Long
    expect ("(")
    x = numeric_expr#
    expect (",")
    y = numeric_expr#
    expect (")")
    If accept&(",") Then
        clr = numeric_expr#
        PSet (x, y), clr
    Else
        PSet (x, y)
    End If
End Sub

Sub randomizer
    If sym = "" Then Randomize Timer Else Randomize numeric_expr#
End Sub

Sub returnstmt
    Dim lin As Integer, offs As Integer
    If stackp < 0 Then Print at_line$; "return without gosub": errors = true: Exit Sub
    lin = gosubstack(stackp)
    offs = gosuboffstack(stackp)
    If tracing Then Print "returning to: "; lin; ": "; offs
    'print "["; curline; "] "; "returning to: "; lin; ": "; offs; " while_sp: "; while_sp
    stackp = stackp - 1
    If offs <= 1 Then Print at_line$; "returnstmt - offs <= 1": errors = true
    Call initgetsym(lin, offs)
End Sub

' SCREEN mode% [,[colorswitch%] [,[activepage%] [,visualpage%]]]
Sub screenstmt
    Screen numeric_expr#
End Sub

' shell [string]
Sub shellstmt
    Dim s As String

    s = ""
    While the_ch <> "" And the_ch <> ":"
        s = s + the_ch
        Call getch
    Wend

    Shell s
    'print "shell: "; s
    Call skiptoeol
End Sub

' sleep [seconds]
Sub sleepstmt
    If is_stmt_end& Then Sleep Else Sleep numeric_expr#
End Sub

' swap v1, v2
Sub swapstmt
    Dim i1 As Integer, i2 As Integer
    Dim symtype1 As Integer, symtype2 As Integer
    Dim sym1 As String, sym2 As String

    sym1 = sym
    symtype1 = symtype
    If symtype = tyident Then
        i1 = getvarindex&(left_side)
    Else
        i1 = getstrindex&(left_side)
    End If

    expect (",")

    sym2 = sym
    symtype2 = symtype

    If symtype = tyident Then
        i2 = getvarindex&(left_side)
    Else
        i2 = getstrindex&(left_side)
    End If

    If symtype1 <> symtype2 Then
        Print at_line$; sym1; " and "; sym2; " are not the same data type": errors = true
        Exit Sub
    End If

    If symtype1 = tyident Then
        Swap numeric_store(i1), numeric_store(i2)
    Else
        Swap string_store(i1), string_store(i2)
    End If
End Sub

' VIEW [[SCREEN] (x1!,y1!)-(x2!,y2!) [,[color%] [,border%]]]
Sub viewstmt
    Dim x1 As Long, y1 As Long, x2 As Long, y2 As Long, clr As Long, border As Long

    expect ("(")
    x1 = numeric_expr#
    expect (",")
    y1 = numeric_expr#
    expect (")")

    expect ("-")

    expect ("(")
    x2 = numeric_expr#
    expect (",")
    y2 = numeric_expr#
    expect (")")

    If accept&(",") Then
        If accept&(",") Then
            border = numeric_expr#
            View (x1, y1)-(x2, y2), , border
        Else
            clr = numeric_expr#
            If accept&(",") Then
                border = numeric_expr#
                View (x1, y1)-(x2, y2), clr, border
            Else
                View (x1, y1)-(x2, y2), clr
            End If
        End If
    Else
        View (x1, y1)-(x2, y2)
    End If
End Sub

Sub whilestmt (first As Integer)
    If first Then
        while_sp = while_sp + 1
        while_line(while_sp) = curline
        while_off(while_sp) = textp
        If Len(sym) > 0 Then while_off(while_sp) = textp - Len(sym) - 1
    End If
    'print "["; curline; "] "; "*while:sym:";sym; " textp:";textp; " =>";mid$(pgm(curline), textp); " while_sp: "; while_sp
    If Not bool_expr& Then
        while_sp = while_sp - 1
        'print "["; curline; "] "; "*wend bool_expr is 0!"; " while_sp: "; while_sp
        Call find_matching_pair("while", "wend")
        Call getsym
    End If
End Sub

Sub wendstmt
    If while_sp <= 0 Then errors = true: Print at_line$; "wend without while": errors = true: Exit Sub
    Call initgetsym(while_line(while_sp), while_off(while_sp))
    If tracing Then Print "["; curline; "] "; "wend"
    whilestmt (false)
End Sub

' do [(while|until) expr][:]
Sub dostmt (first As Integer)
    If first Then
        do_sp = do_sp + 1
        do_loop(do_sp).lline = curline

        do_loop(do_sp).loff = textp
        If Len(sym) > 0 Then do_loop(do_sp).loff = textp - Len(sym) - 1
        'print "*do:"; "sym:"; sym; " textp:";textp; "=>";mid$(pgm(curline), textp - len(sym))
    End If

    If sym = "while" Then
        Call getsym
        If Not bool_expr& Then
            do_sp = do_sp - 1
            Call find_matching_pair("do", "loop")
            Call getsym
        End If
    ElseIf sym = "until" Then
        Call getsym
        If bool_expr& Then
            do_sp = do_sp - 1
            Call find_matching_pair("do", "loop")
            Call getsym
        End If
    End If
End Sub

' loop [(while|until) expr]
Sub loopstmt
    If do_sp <= 0 Then errors = true: Print at_line$; "loop without do": errors = true: Exit Sub

    If sym = "while" Then
        Call getsym
        If Not bool_expr& Then
            do_sp = do_sp - 1
            Exit Sub
        End If
    ElseIf sym = "until" Then
        Call getsym
        If bool_expr& Then
            do_sp = do_sp - 1
            Exit Sub
        End If
    End If

    Call initgetsym(do_loop(do_sp).lline, do_loop(do_sp).loff)
    'print "loop line:"; curline; "off:"; do_loop(do_sp).loff; "==>"; pgm(curline)
    dostmt (false)
End Sub

' width , height
' width width
' width width, height
Sub widthstmt
    Dim w As Integer

    If accept&(",") Then
        Width , numeric_expr#
    Else
        w = numeric_expr#
        If accept&(",") Then
            Width w, numeric_expr#
        Else
            Width w
        End If
    End If
End Sub

' window [ [ screen] (x1!, y1!) - (x2!, y2!)]
Sub windowstmt
    Dim x1 As Single, y1 As Single, x2 As Single, y2 As Single

    If sym = "" Then Window: Exit Sub

    expect ("(")
    x1 = numeric_expr#
    expect (",")
    y1 = numeric_expr#
    expect (")")

    expect ("-")

    expect ("(")
    x2 = numeric_expr#
    expect (",")
    y2 = numeric_expr#
    expect (")")

    Window (x1, y1)-(x2, y2)
End Sub

'------------------------------------------------------------------------
'  Various helper routines
'------------------------------------------------------------------------

Sub skip_exit
    Call getsym
    If sym = "do" Or sym = "while" Or sym = "for" Then
        Call getsym
    End If
End Sub

Sub find_matching_pair (s1 As String, s2 As String)
    Dim level As Integer
    Dim more As Integer
    Dim have_sym As Integer

    level = 1
    more = true
    have_sym = false

    endif_count = 0: wend_count = 0: next_count = 0: loop_count = 0

    Do
        If sym = "exit" Then
            Call skip_exit
        ElseIf Not have_sym Then
            Call getsym
            If sym = "exit" Then Call skip_exit
        End If

        have_sym = false
        'print at_line$; "matching, level"; level; "sym=>"; sym
        'if isalpha&(mid$(sym, 1, 1)) then print "fm: level: sym: "; level; ": '"; sym; "'  "; mid$(thelin, textp, 40)

        Select Case sym
            Case s1: level = level + 1
            Case s2: level = level - 1
        End Select

        If level = 0 Then Exit Do

        Select Case sym
            Case "if" ' need to only do "if" Case If multiline If
                Do
                    Call getsym
                Loop Until sym = "then" Or sym = ""
                Call getsym ' skip the "then"
                ' if nothing past "then", it is a multiline If
                If sym = "" Then endif_count = endif_count - 1
            Case "endif": endif_count = endif_count + 1
            Case "while": wend_count = wend_count - 1
            Case "wend": wend_count = wend_count + 1
            Case "for": next_count = next_count - 1
            Case "next": next_count = next_count + 1
            Case "do"
                loop_count = loop_count - 1
                If sym = "while" Then Call getsym
            Case "loop"
                loop_count = loop_count + 1
                If sym = "while" Then Call getsym
        End Select

        If sym = "" Then
            If Not more Then Exit Sub
            While sym = "" And curline > 0 And curline < pgmsize
                Call initgetsym(curline + 1, 1)
            Wend
            If sym = "" Then
                Print at_line$; "Cannot find matching: "; s2: errors = true
                Exit Sub
            End If
            have_sym = true
        End If
    Loop
End Sub

' find matching elseif/else/endif
Function find_matching_else$
    Dim level As Integer

    find_matching_else$ = ""
    level = 0
    Do
        Call initgetsym(curline + 1, 1)
        'print "find_matching_else: "; curline; " sym: "; sym; " level: "; level; "textp: "; textp; " line:"; thelin
        If curline >= pgmsize Then Print "searching for endif, found eof": errors = true: Exit Do
        If is_multi_line_if& Then
            level = level + 1
        ElseIf level = 0 And (sym = "elseif" Or sym = "else" Or sym = "endif") Then
            find_matching_else = sym: Exit Do
        ElseIf level > 0 And sym = "endif" Then
            level = level - 1
        ElseIf errors Then
            Exit Do
        End If
    Loop
End Function

Sub find_matching_sline_if
    Dim level As Integer
    level = 1
    Do
        'print "find_matching_sline_if level: "; level; " sym: "; sym
        If sym = "if" Then
            level = level + 1
        ElseIf sym = "else" Then
            level = level - 1
        End If
        If level = 0 Or sym = "" Then Exit Do
        Call getsym
    Loop
End Sub

Function is_multi_line_if&
    is_multi_line_if& = false

    If sym = "if" Then
        ' is it single or multi line "if" - ignore single line If's
        Do
            Call getsym
            If sym = "" Then Print at_line$; "if missing then": errors = true: Exit Do
            If sym = "then" Then
                Call getsym
                If sym = "" Then
                    ' multi line "if"
                    is_multi_line_if& = true
                End If
                Exit Do
            End If
        Loop
    End If
End Function

Function accept& (s As String)
    accept& = false
    If sym = s Then accept& = true: Call getsym
End Function

Sub expect (s As String)
    If Not accept&(s) Then Print at_line$; "expecting "; s; " but found "; sym: errors = true
End Sub

Function is_stmt_end&
    is_stmt_end& = sym = "" Or sym = ":"
End Function

Sub validlinenum (n As Integer)
    If n > 0 And n <= pgmsize Then Exit Sub
    Print at_line$; "line number out of range:"; the_num: errors = true
End Sub

Function storeline&
    'print "storeline"
    storeline& = false
    Call initgetsym(0, 1)
    If symtype = tynum Then
        validlinenum (Int(the_num))
        pgm(the_num) = Mid$(pgm(0), textp, Len(pgm(0)) - textp + 1)
        storeline& = true
    End If
End Function

Sub clearprog
    Dim i As Integer
    For i = 1 To pgmsize
        pgm(i) = ""
    Next
End Sub

Sub gotoline (target As Integer)
    validlinenum (target)
    Call initgetsym(target, 1)
End Sub

'------------------------------------------------------------------------
'------[QB64 specific functions]-----------------------------------------
'------------------------------------------------------------------------

' _atan2(y, x)
Function atan2fun#
    Dim x As Double, y As Double

    expect ("(")
    y = numeric_expr#
    expect (",")
    x = numeric_expr#
    expect (")")
    atan2fun# = _Atan2(y, x)
End Function

Sub filesstmt
    Dim s As String
    s = ""
    If symtype = tystring Or symtype = tystrident Then s = strexpression$
    Files s
End Sub

' freeimage [image]&
Sub freeimage
    If sym <> "" And sym <> ":" Then
        _FreeImage numeric_expr#
    Else
        _FreeImage
    End If
End Sub

Sub limitstmt
    _Limit numeric_expr#
End Sub

' ([start], haystack, needle)
Function instrrevfun&
    Dim i As Integer, haystack As String, needle As String

    expect ("(")
    i = 0
    If symtype = tynum Or symtype = tyident Then
        i = numeric_expr#
        expect (",")
    End If
    haystack = strexpression$
    expect (",")
    needle = strexpression$
    expect (")")
    instrrevfun& = _InStrRev(i, haystack, needle)
End Function

Sub titlestmt
    _Title strexpression$
End Sub

' _newimage(width&, height&[, {0|1|2|7|8|9|10|11|12|13|256|32}])
Function newimagefun&
    Dim w As Long, h As Long, mode As Long

    expect ("(")
    w = numeric_expr#
    expect (",")
    h = numeric_expr#
    If accept&(",") Then
        mode = numeric_expr#
        newimagefun& = _NewImage(w, h, mode)
    Else
        newimagefun& = _NewImage(w, h)
    End If
    expect (")")
End Function

'colorIndex~& = _RGB(red&, green&, blue&[, imageHandle&])
Function rgbfun~&
    Dim r As Long, g As Long, b As Long, h As Long
    expect ("(")
    r = numeric_expr#
    expect (",")
    g = numeric_expr#
    expect (",")
    b = numeric_expr#
    If accept&(",") Then
        h = numeric_expr#
        rgbfun = _RGB(r, g, b, h)
    Else
        rgbfun = _RGB(r, g, b)
    End If
    expect (")")
End Function

'color32value~& = _RGB32(red&, green&, blue&, alpha&)
'color32value~& = _RGB32(red&, green&, blue&)
'color32value~& = _RGB32(intensity&, alpha&)
'color32value~& = _RGB32(intensity&)
Function rgb32fun~&
    Dim r As Long, g As Long, b As Long, a As Long
    expect ("(")
    r = numeric_expr#
    If accept&(")") Then
        rgb32fun = _RGB32(r)
    Else
        expect (",")
        g = numeric_expr#
        If accept&(")") Then
            rgb32fun = _RGB32(r, g)
        Else
            expect (",")
            b = numeric_expr#
            If accept&(")") Then
                rgb32fun = _RGB32(r, g, b)
            Else
                expect (",")
                a = numeric_expr#
                rgb32fun = _RGB32(r, g, b, a)
                expect (")")
            End If
        End If
    End If
End Function

'_RGBA(red&, green&, blue&, alpha&[, imageHandle&])
Function rgbafun~&
    Dim r As Long, g As Long, b As Long, a As Long, h As Long

    expect ("(")
    r = numeric_expr#
    expect (",")

    g = numeric_expr#
    expect (",")

    b = numeric_expr#
    expect (",")

    a = numeric_expr#

    If accept&(",") Then
        h = numeric_expr#
        rgbafun = _RGBA(r, g, b, a, h)
    Else
        rgbafun = _RGBA(r, g, b, a)
    End If

    expect (")")
End Function

'color32value~& = _RGBA32(red&, green&, blue&, alpha&)
Function rgba32fun~&
    Dim r As Long, g As Long, b As Long, a As Long

    expect ("(")
    r = numeric_expr#
    expect (",")

    g = numeric_expr#
    expect (",")

    b = numeric_expr#
    expect (",")

    a = numeric_expr#

    rgba32fun = _RGBA32(r, g, b, a)

    expect (")")
End Function

'_PRINTSTRING(column, row), textExpression$[, imageHandle&]
Sub printstringstmt
    Dim c As Integer, r As Integer, ex As String

    expect ("(")
    c = numeric_expr#
    expect (",")
    r = numeric_expr#
    expect (")")
    expect (",")
    ex = strexpression$
    If accept&(",") Then
        _PrintString (c, r), ex, numeric_expr#
    Else
        _PrintString (c, r), ex
    End If
End Sub

' _SCREENMOVE {column&, row&|_MIDDLE}
Sub screenmovestmt
    Dim c As Long, r As Long

    'print "screenmovestmt:"; sym
    If UCase$(sym) = "_MIDDLE" Then
        _ScreenMove _Middle
        Call getsym 'and skip over _middle
    Else
        c = numeric_expr#
        expect (",")
        r = numeric_expr#
        _ScreenMove c, r
    End If
End Sub

' sound frequence, duration
Sub soundstmt
    Dim f As Long, d As Single

    f = numeric_expr#
    expect (",")
    d = numeric_expr#
    Sound f, d
End Sub

'------------------------------------------------------------------------
'  various functions called from primary
'------------------------------------------------------------------------

Function sinh# (z As Double)
    sinh = (e ^ z - e ^ (-z)) / 2
End Function

Function tanh# (z As Double)
    tanh = (e ^ (2 * z) - 1) / (e ^ (2 * z) + 1)
End Function

Function acosh# (z As Double)
    acosh = Log(z + Sqr(z + 1) * Sqr(z - 1))
End Function

Function acoth# (z As Double)
    acoth = .5 * (Log(1 + 1 / z) - Log(1 - 1 / z))
End Function

Function acsch# (z As Double)
    acsch = Log(Sqr(1 + z ^ (-2)) + z ^ (-1))
End Function

Function asech# (z As Double)
    asech = Log(Sqr(z ^ (-1) - 1) * Sqr(z ^ (-1) + 1) + z ^ (-1))
End Function

Function asin2# (i As Double)
    If i = -1 Then
        asin2 = -halfpi
    ElseIf i = 1 Then
        asin2 = halfpi
    Else
        asin2 = Atn(i / Sqr(1 - i * i))
    End If
End Function

Function asinh# (z As Double)
    asinh = Log(z + Sqr(1 + z ^ 2))
End Function

Function atanh# (z As Double)
    atanh = .5 * (Log(1 + z) - Log(1 - z))
End Function

Function cosh# (z As Double)
    cosh = (e ^ z + e ^ (-z)) / 2
End Function

Function shlf# (x As Double, n As Double)
    shlf# = x
    If n >= 0 Then shlf# = x * (2 ^ n)
End Function

Function shrf# (x As Double, n As Double)
    shrf# = x
    If n >= 0 Then shrf# = x \ (2 ^ n)
End Function

' ([start,] haystack, needle)
Function instrfun&
    Dim i As Integer, haystack As String, needle As String

    expect ("(")
    i = 1
    If symtype = tynum Or symtype = tyident Then
        i = numeric_expr#
        expect (",")
    End If
    haystack = strexpression$
    expect (",")
    needle = strexpression$
    expect (")")
    instrfun& = InStr(i, haystack, needle)
End Function

' mid$(s$, start [, end])
Function midfun$
    Dim i As String, x As Integer, y As Integer

    expect ("(")
    i = strexpression$
    expect (",")
    x = numeric_expr#
    If accept&(",") Then
        y = numeric_expr#
        midfun$ = Mid$(i, x, y)
    Else
        midfun$ = Mid$(i, x)
    End If
    expect (")")
End Function

' lpad$(s$, padded_len [, pad_string$])
Function lpadfun$
    Dim s As String, pad_string As String, padded_len As Integer
    expect ("(")
    s = strexpression$
    expect (",")
    padded_len = numeric_expr#
    pad_string = " "
    If accept&(",") Then
        pad_string = strexpression$
    End If
    expect (")")

    lpadfun$ = s
    If Len(s) > padded_len Then
        lpadfun$ = Mid$(s, 1, padded_len)
    Else
        lpadfun$ = String$(padded_len - Len(s), Mid$(pad_string, 1, 1)) + s
    End If
End Function

' result = peek(string)
Function peekfun# (s As String)
    Select Case s
        Case "for index": peekfun# = loopp
        Case "do index": peekfun# = do_sp
        Case "while index": peekfun# = while_sp
        Case "if index": peekfun# = if_sp
        Case "gosub index": peekfun# = stackp
        Case "numeric var total": peekfun# = num_store_max
        Case "string var total": peekfun# = str_store_max
        Case "variables total": peekfun# = var_names_max
        Case Else: peekfun# = -1
    End Select
End Function

'result = Point( coord_x, coord_y [,buffer] )
'result = Point( function_index )
Function pointfun#
    Dim x As Integer

    expect ("(")
    x = numeric_expr#
    If accept&(",") Then
        pointfun# = Point(x, numeric_expr#)
    Else
        pointfun# = Point(x)
    End If
    expect (")")
End Function

Function posfun#
    expect ("(")
    posfun# = Pos(numeric_expr#)
    expect (")")
End Function

' rpad$(s$, padded_len [, pad_string$])
Function rpadfun$
    Dim s As String, pad_string As String, padded_len As Integer
    expect ("(")
    s = strexpression$
    expect (",")
    padded_len = numeric_expr#
    pad_string = " "
    If accept&(",") Then
        pad_string = strexpression$
    End If
    expect (")")

    rpadfun$ = s
    If Len(s) > padded_len Then
        rpadfun$ = Mid$(s, 1, padded_len)
    Else
        rpadfun$ = String$(padded_len - Len(s), Mid$(pad_string, 1, 1)) + s
    End If
End Function

' replace$(haystack$, needle$ [, newst$])
Function replacefun$
    Dim haystack As String, needle As String, newst As String, start As Integer, p As Integer
    expect ("(")
    haystack = strexpression$
    expect (",")
    needle = strexpression$

    newst = ""
    If accept&(",") Then
        newst = strexpression$
    End If
    expect (")")

    start = 1
    Do
        p = InStr(start, haystack, needle)
        If p = 0 Then Exit Do
        haystack = Mid$(haystack, 1, p - 1) + newst + Mid$(haystack, p + Len(needle))
        start = p + Len(newst) + 1
    Loop

    replacefun$ = haystack
End Function

' ubound(array-name)
Function uboundfun&
    Dim i As Long
    expect ("(")
    i = find_vname(sym)
    If i = 0 Then
        Print at_line$; "ubound: not an array: "; sym: errors = true
    Else
        uboundfun& = var_names(i).hi_bnd
    End If
    Call getsym
    expect (")")
End Function

' screen(row, col)
Function screenfun&
    Dim row As Long, col As Long
    expect ("(")
    row = numeric_expr#
    expect (",")
    col = numeric_expr#
    expect (")")
    screenfun& = Screen(row, col)
End Function

'------------------------------------------------------------------------
' expression parser
'------------------------------------------------------------------------

Function binary_prec& (op As String)
    Select Case op
        Case "^": binary_prec& = 14
        Case "*", "/": binary_prec& = 12
        Case "\": binary_prec& = 11
        Case "mod": binary_prec& = 10
        Case "+", "-": binary_prec& = 9
        Case ">>", "<<", "shl", "shr": binary_prec& = 8
        Case "=", "<>", "<", ">", "<=", ">=": binary_prec& = 7
        Case "and": binary_prec& = 5
        Case "or": binary_prec& = 4
        Case "xor": binary_prec& = 3
        Case "eqv": binary_prec& = 2
        Case "imp": binary_prec& = 1
        Case Else: binary_prec& = 0
    End Select
End Function

Function strfactor$
    Dim x As Integer
    Dim s As String

    Select Case sym
        Case "chr$": Call getsym: expect ("("): strfactor$ = Chr$(numeric_expr#): expect (")")
        Case "command$": Call getsym: strfactor$ = Command$
        Case "date$": Call getsym: strfactor$ = Date$
        Case "environ$": Call getsym: expect ("("): strfactor$ = Environ$(strexpression$): expect (")")
        Case "hex$": Call getsym: expect ("("): strfactor$ = Hex$(numeric_expr#): expect (")")
        Case "inkey$": Call getsym: strfactor$ = InKey$
        Case "lcase$": Call getsym: expect ("("): strfactor$ = LCase$(strexpression$): expect (")")
        Case "left$"
            Call getsym:
            expect ("(")
            s = strexpression$
            expect (",")
            x = numeric_expr#
            strfactor$ = Left$(s, x)
            expect (")")
        Case "lpad$": Call getsym: strfactor$ = lpadfun$
        Case "ltrim$": Call getsym: expect ("("): strfactor$ = LTrim$(strexpression$): expect (")")
        Case "mid$": Call getsym: strfactor$ = midfun$
        Case "mki$": Call getsym: expect ("("): strfactor$ = MKI$(numeric_expr#): expect (")")
        Case "oct$": Call getsym: expect ("("): strfactor$ = Oct$(numeric_expr#): expect (")")
        Case "replace$": Call getsym: strfactor$ = replacefun$
        Case "right$"
            Call getsym
            expect ("(")
            s = strexpression$
            expect (",")
            x = numeric_expr#
            strfactor$ = Right$(s, x)
            expect (")")
        Case "rpad$": Call getsym: strfactor$ = rpadfun$
        Case "rtrim$": Call getsym: expect ("("): strfactor$ = RTrim$(strexpression$): expect (")")
        Case "space$"
            Call getsym
            expect ("(")
            strfactor$ = Space$(numeric_expr#)
            expect (")")
        Case "str$": Call getsym: expect ("("): strfactor$ = Str$(numeric_expr#): expect (")")
        Case "string$"
            Call getsym ' string$(n [, strexpr])
            expect ("(")
            x = numeric_expr#
            expect (",")
            If symtype = tystring Or symtype = tystrident Then
                strfactor$ = String$(x, strexpression$)
            Else
                strfactor$ = String$(x, numeric_expr#)
            End If
            expect (")")
        Case "time$": Call getsym: strfactor$ = Time$
        Case "trim$": Call getsym: expect ("("): strfactor$ = LTrim$(RTrim$(strexpression$)): expect (")")
        Case "ucase$": Call getsym: expect ("("): strfactor$ = UCase$(strexpression$): expect (")")

        Case "_clipboard$": Call getsym: strfactor$ = _Clipboard$
        Case "_cwd$": Call getsym: strfactor$ = _CWD$
        Case "_os$": Call getsym: strfactor$ = _OS$
        Case "_startdir$": Call getsym: strfactor$ = _StartDir$
        Case "_title$": Call getsym: strfactor$ = _Title$
        Case "_trim$": Call getsym: expect ("("): strfactor$ = LTrim$(RTrim$(strexpression$)): expect (")")

        Case Else
            If symtype = tystring Then
                strfactor$ = Mid$(sym, 2, Len(sym) - 1)
                Call getsym
            ElseIf symtype = tystrident Then
                If peek_ch$ = "(" Then
                    strfactor$ = get_string_array_value$
                Else
                    strfactor$ = string_store(getstrindex&(right_side))
                End If
            Else
                Print at_line$; "In strfactor, expecting an operand, found: "; sym; " symtype is: "; symtype: errors = true
                Call getsym
            End If
    End Select
End Function

Function frac# (n As Double)
    frac# = n - Fix(n)
End Function

Function primary#
    Dim i As Double

    Select Case sym
        Case "-": Call getsym: primary# = -numeric_expr2#(unaryminus_prec)
        Case "+": Call getsym: primary# = numeric_expr2#(unaryplus_prec)
        Case "not": Call getsym: primary# = Not numeric_expr2#(unarynot_prec)
        Case "abs": Call getsym: expect ("("): primary# = Abs(numeric_expr#): expect (")")
        Case "acos": Call getsym: expect ("("): primary# = halfpi - asin2(numeric_expr#): expect (")")
        Case "acosh": Call getsym: expect ("("): primary# = acosh(numeric_expr#): expect (")")
        Case "acot": Call getsym: expect ("("): primary# = halfpi - Atn(numeric_expr#): expect (")")
        Case "acoth": Call getsym: expect ("("): primary# = acoth(numeric_expr#): expect (")")
        Case "acsc": Call getsym: expect ("("): primary# = asin2(1 / numeric_expr#): expect (")")
        Case "acsch": Call getsym: expect ("("): primary# = acsch(numeric_expr#): expect (")")
        Case "asc": Call getsym: expect ("("): primary# = Asc(strexpression$): expect (")")
        Case "asec": Call getsym: expect ("("): primary# = halfpi - asin2(1 / numeric_expr#): expect (")")
        Case "asech": Call getsym: expect ("("): primary# = asech(numeric_expr#): expect (")")
        Case "asin": Call getsym: expect ("("): primary# = asin2(numeric_expr#): expect (")")
        Case "asinh": Call getsym: expect ("("): primary# = asinh(numeric_expr#): expect (")")
        Case "atanh": Call getsym: expect ("("): primary# = atanh(numeric_expr#): expect (")")
        Case "atn", "atan": Call getsym: expect ("("): primary# = Atn(numeric_expr#): expect (")")
        Case "cdbl": Call getsym: expect ("("): primary# = CDbl(numeric_expr#): expect (")")
        Case "cint": Call getsym: expect ("("): primary# = CInt(numeric_expr#): expect (")")
        Case "clng": Call getsym: expect ("("): primary# = CLng(numeric_expr#): expect (")")
        Case "cos": Call getsym: expect ("("): primary# = Cos(numeric_expr#): expect (")")
        Case "cosh": Call getsym: expect ("("): primary# = cosh(numeric_expr#): expect (")")
        Case "cot": Call getsym: expect ("("): primary# = 1 / Tan(numeric_expr#): expect (")")
        Case "coth": Call getsym: expect ("("): primary# = 1 / tanh(numeric_expr#): expect (")")
        Case "csc": Call getsym: expect ("("): primary# = 1 / Sin(numeric_expr#): expect (")")
        Case "csch": Call getsym: expect ("("): primary# = 1 / sinh(numeric_expr#): expect (")")
        Case "csng": Call getsym: expect ("("): primary# = CSng(numeric_expr#): expect (")")
        Case "csrlin": Call getsym: primary# = CsrLin
        Case "cvd": Call getsym: expect ("("): primary# = CVD(strexpression$): expect (")")
        Case "cvi": Call getsym: expect ("("): primary# = CVI(strexpression$): expect (")")
        Case "exp": Call getsym: expect ("("): primary# = Exp(numeric_expr#): expect (")")
        Case "false": Call getsym: primary# = false
        Case "frac": Call getsym: expect ("("): primary# = frac#(numeric_expr#): expect (")")
        Case "fix": Call getsym: expect ("("): primary# = Fix(numeric_expr#): expect (")")
        Case "instr": Call getsym: primary# = instrfun&
        Case "int": Call getsym: expect ("("): primary# = Int(numeric_expr#): expect (")")
        Case "len": Call getsym: expect ("("): primary# = Len(strexpression$): expect (")")
        Case "ln": Call getsym: expect ("("): primary# = Log(numeric_expr#): expect (")")
        Case "log"
            Call getsym:
            expect ("(")
            i = Log(numeric_expr#)
            If accept&(",") Then
                primary# = i / Log(numeric_expr#)
            Else
                primary# = i
            End If
            expect (")")
        Case "log10": Call getsym: expect ("("): primary# = Log(numeric_expr#) / Log(10): expect (")")
        Case "peek": Call getsym: expect ("("): primary# = peekfun#(strexpression$): expect (")")
        Case "point": Call getsym: primary# = pointfun#
        Case "pos": Call getsym: primary# = posfun#
        Case "rnd"
            Call getsym:
            If accept&("(") Then
                primary# = Rnd(numeric_expr#)
                expect (")")
            Else
                primary# = Rnd
            End If
        Case "screen": Call getsym: primary# = screenfun&
        Case "sec": Call getsym: expect ("("): primary# = 1 / Cos(numeric_expr#): expect (")")
        Case "sech": Call getsym: expect ("("): primary# = 1 / cosh(numeric_expr#): expect (")")
        Case "sgn": Call getsym: expect ("("): primary# = Sgn(numeric_expr#): expect (")")
        Case "sin": Call getsym: expect ("("): primary# = Sin(numeric_expr#): expect (")")
        Case "sinh": Call getsym: expect ("("): primary# = sinh(numeric_expr#): expect (")")
        Case "sqr", "sqrt": Call getsym: expect ("("): primary# = Sqr(numeric_expr#): expect (")")
        Case "tan": Call getsym: expect ("("): primary# = Tan(numeric_expr#): expect (")")
        Case "tanh": Call getsym: expect ("("): primary# = tanh(numeric_expr#): expect (")")
        Case "timer": Call getsym: primary# = Timer
        Case "true": Call getsym: primary# = true
        Case "ubound": Call getsym: primary# = uboundfun
        Case "val": Call getsym: expect ("("): primary# = Val(strexpression$): expect (")")

        Case "_atan2": Call getsym: primary# = atan2fun#
        Case "_ceil": Call getsym: expect ("("): primary# = _Ceil(numeric_expr#): expect (")")
        Case "_d2g": Call getsym: expect ("("): primary# = _D2G(numeric_expr#): expect (")")
        Case "_d2r": Call getsym: expect ("("): primary# = _D2R(numeric_expr#): expect (")")
        Case "_fontwidth": Call getsym: primary# = _FontWidth
        Case "_fontheight": Call getsym: primary# = _FontHeight
        Case "_g2d": Call getsym: expect ("("): primary# = _G2D(numeric_expr#): expect (")")
        Case "_g2r": Call getsym: expect ("("): primary# = _G2R(numeric_expr#): expect (")")
        Case "_height": Call getsym: primary# = _Height
        Case "_instrrev": Call getsym: primary# = instrrevfun&
        Case "_keydown": Call getsym: expect ("("): primary# = _KeyDown(numeric_expr#): expect (")")
        Case "_keyhit": Call getsym: primary# = _KeyHit
        Case "_mousebutton": Call getsym: expect ("("): primary# = _MouseButton(numeric_expr#): expect (")")
        Case "_mouseinput": Call getsym: primary# = _MouseInput
        Case "_mousex": Call getsym: primary# = _MouseX
        Case "_mousey": Call getsym: primary# = _MouseY
        Case "_newimage": Call getsym: primary# = newimagefun&
        Case "_pi": Call getsym: primary# = _Pi
        Case "_r2d": Call getsym: expect ("("): primary# = _R2D(numeric_expr#): expect (")")
        Case "_r2g": Call getsym: expect ("("): primary# = _R2G(numeric_expr#): expect (")")
        Case "_rgb": Call getsym: primary# = rgbfun
        Case "_rgba": Call getsym: primary# = rgbafun
        Case "_rgba32": Call getsym: primary# = rgba32fun
        Case "_rgb32": Call getsym: primary# = rgb32fun

        Case "_round": Call getsym: expect ("("): primary# = _Round(numeric_expr#): expect (")")
        Case "_width": Call getsym: primary# = _Width

        Case Else
            If Left$(sym, 1) = "_" Then
                Print "Unknown function: "; sym: errors = true: Call getsym
            ElseIf symtype = tynum Then
                primary# = the_num
                Call getsym
            ElseIf symtype = tyident Then
                If peek_ch$ = "(" Then
                    primary# = get_numeric_array_value#
                Else
                    primary# = numeric_store(getvarindex&(right_side))
                End If
            Else
                Print at_line$; "In primary, expecting an operand, found: "; sym; " symtype is: "; symtype: errors = true
                Call getsym
            End If
    End Select
End Function

Function strexpression$
    Dim s As String
    s = strfactor$
    While accept&("+")
        s = s + strfactor$
    Wend
    strexpression$ = s
End Function

'-------------------------------------------------------------------------------------------------

Sub push_str (s As String)
    str_st_ndx = str_st_ndx + 1
    str_stack(str_st_ndx) = s
End Sub

Sub push_num (n As Double)
    num_st_ndx = num_st_ndx + 1
    num_stack(num_st_ndx) = n
End Sub

Function pop_str$
    pop_str = str_stack(str_st_ndx)
    str_st_ndx = str_st_ndx - 1
End Function

Function pop_num#
    pop_num = num_stack(num_st_ndx)
    num_st_ndx = num_st_ndx - 1
End Function

Function evalstrexpr& (op As String)
    Dim s As String, s2 As String, n As Double

    s2 = pop_str$
    s = pop_str$
    Select Case op
        Case "=": n = s = s2
        Case "<>": n = s <> s2
        Case "<": n = s < s2
        Case ">": n = s > s2
        Case "<=": n = s <= s2
        Case ">=": n = s >= s2
        Case Else
            Print at_line$; "In expr, expecting a string operator, found: "; op; " symtype is: "; symtype: errors = true
            Call getsym
    End Select
    push_num (n)
    evalstrexpr& = tynum
End Function

Function evalnumericexpr& (op As String)
    Dim n As Double, n2 As Double

    n2 = pop_num#
    n = pop_num#
    Select Case op
        Case "^": n = n ^ n2
        Case "*": n = n * n2
        Case "/": If n2 = 0 Then Print at_line$; "division by 0": errors = true Else n = n / n2
        Case "\": If n2 = 0 Then Print at_line$; "division by 0": errors = true Else n = n \ n2
        Case "mod": If n2 = 0 Then Print at_line$; "division by 0": errors = true Else n = n Mod n2
        Case "+": n = n + n2
        Case "-": n = n - n2
        Case "=": n = n = n2
        Case "<>": n = n <> n2
        Case "<": n = n < n2
        Case ">": n = n > n2
        Case "<=": n = n <= n2
        Case ">=": n = n >= n2
        Case "and": n = n And n2
        Case "or": n = n Or n2
        Case "xor": n = n Xor n2
        Case "eqv": n = n Eqv n2
        Case "imp": n = n Imp n2
        Case "shl", "<<": n = shlf#(n, n2)
        Case "shr", ">>": n = shrf#(n, n2)
        Case Else
            Print at_line$; "In expr, expecting a numeric operator, found: "; op; " symtype is: "; symtype: errors = true
            Call getsym
    End Select
    push_num (n)
    evalnumericexpr& = tynum
End Function

' return the type of expression, either string or numeric; result is on the stack
Function any_expr& (p As Integer)
    Dim left_type As Integer

    ' we need to decide which primary to call - numeric or string
    ' leading parens don't tell us which primary, so just do recursive call
    If accept&("(") Then
        left_type = any_expr&(0)
        expect (")")
    ElseIf symtype = tystring Or symtype = tystrident Then
        push_str (strexpression$)
        left_type = tystring
    ElseIf symtype = tynum Or symtype = tyident Or sym = "-" Or sym = "+" Or sym = "not" Then
        push_num (primary#)
        left_type = tynum
    ElseIf sym = "" Then
        Print at_line$; "In expr, unexpected end-of-line found: "; pgm(curline): errors = true
    Else
        Print at_line$; "In expr, expecting an expr, found: "; sym; " symtype is: "; symtype; " - near column: "; textp
        errors = true
        Call getsym
    End If

    Do ' while binary operator and precedence(sym) >= p
        Dim op As String
        Dim right_type As Integer, prec As Integer

        prec = binary_prec&(sym)
        If prec = 0 Or prec < p Then Exit Do

        op = sym

        Call getsym

        ' all operators are left associative in qbasic
        prec = prec + 1

        right_type = any_expr&(prec)

        If left_type = tystring And right_type = tystring Then
            left_type = evalstrexpr&(op)
        ElseIf left_type = tynum And right_type = tynum Then
            left_type = evalnumericexpr&(op)
        Else
            Print at_line$; "type missmatch in expr - left_type:"; left_type; " right_type:"; right_type: errors = true
            Call getsym
        End If
    Loop
    any_expr& = left_type
End Function

Function numeric_expr2# (p As Integer)
    If any_expr&(p) = tynum Then
        numeric_expr2# = pop_num#
    Else
        Print at_line$; "numeric expression expected": errors = true
    End If
End Function

' process and return a numeric expression
Function numeric_expr#
    numeric_expr# = numeric_expr2#(0)
End Function

Function bool_expr&
    bool_expr& = (numeric_expr# <> 0)
End Function

'------------------------------------------------------------------------
'  scanner
'------------------------------------------------------------------------
Sub init_scanner
    Dim i As Integer

    For i = 0 To 255
        ctype_arr(i) = ct_unknown
    Next

    ' alpha
    For i = Asc("a") To Asc("z")
        ctype_arr(i) = ct_alpha
    Next
    For i = Asc("A") To Asc("Z")
        ctype_arr(i) = ct_alpha
    Next
    ctype_arr(Asc("_")) = ct_alpha

    ' num
    For i = Asc("0") To Asc("9")
        ctype_arr(i) = ct_digit
    Next

    ctype_arr(Asc(".")) = ct_period
    ctype_arr(Asc(",")) = ct_punc1
    ctype_arr(Asc(";")) = ct_punc1
    ctype_arr(Asc("=")) = ct_punc1
    ctype_arr(Asc("+")) = ct_punc1
    ctype_arr(Asc("-")) = ct_punc1
    ctype_arr(Asc("*")) = ct_punc1
    ctype_arr(Asc("/")) = ct_punc1
    ctype_arr(Asc("\")) = ct_punc1
    ctype_arr(Asc("^")) = ct_punc1
    ctype_arr(Asc("(")) = ct_punc1
    ctype_arr(Asc(")")) = ct_punc1
    ctype_arr(Asc("?")) = ct_punc1
    ctype_arr(Asc(":")) = ct_punc1
    ctype_arr(Asc("<")) = ct_lt
    ctype_arr(Asc(">")) = ct_gt
    ctype_arr(Asc("&")) = ct_amp
    ctype_arr(Asc(Chr$(34))) = ct_dquote
    ctype_arr(Asc(Chr$(39))) = ct_squote
End Sub

Function peek_ch$
    peek_ch$ = Left$(LTrim$(the_ch) + LTrim$(Mid$(pgm(curline), textp)), 1)
End Function

' other code relies on textp always being incremented; so do it even on EOL
Sub getch
    the_ch = ""
    If textp <= Len(thelin) Then
        the_ch = Mid$(thelin, textp, 1)
    End If
    textp = textp + 1
    'print "getch: textp: "; textp; " the_ch: "; the_ch; " thelin: "; thelin
End Sub

Sub readident
    sym = ""
    Do While the_ch <> ""
        If ctype_arr(Asc(the_ch)) <> ct_alpha And ctype_arr(Asc(the_ch)) <> ct_digit Then Exit Do
        sym = sym + LCase$(the_ch)
        getch
    Loop

    symtype = tyident

    Select Case the_ch
        Case "%", "&", "!", "#": sym = sym + the_ch: getch ' just ignore
        Case "$": symtype = tystrident: sym = sym + the_ch: getch ' string
            ' see if we have "end if", If so, convert to "endif"
        Case " "
            If sym = "end" Then
                If LCase$(Mid$(thelin, textp, 2)) = "if" Then
                    sym = "endif"
                    getch ' skip " "
                    getch ' skip "i"
                    getch ' skip "f"
                End If
            End If
    End Select
End Sub

Sub readnumber
    sym = ""
    Do While the_ch <> ""
        If ctype_arr(Asc(the_ch)) <> ct_digit Then Exit Do
        sym = sym + the_ch
        getch
    Loop
    If the_ch = "." Then
        sym = sym + the_ch
        getch
        Do While the_ch <> ""
            If ctype_arr(Asc(the_ch)) <> ct_digit Then Exit Do
            sym = sym + the_ch
            getch
        Loop
    End If
    If LCase$(the_ch) = "e" Then
        sym = sym + "e"
        getch
        If the_ch = "+" Or the_ch = "-" Then sym = sym + the_ch: getch
        Do While the_ch <> ""
            If ctype_arr(Asc(the_ch)) <> ct_digit Then Exit Do
            sym = sym + the_ch
            getch
        Loop
    End If
    the_num = Val(sym)
    symtype = tynum
End Sub

' on entry pointing to 'h'
Sub readhex
    sym = "&h"
    getch ' skip the 'h'
    Do While the_ch <> ""
        If ctype_arr(Asc(the_ch)) <> ct_digit And InStr("abcdefABCDEF", the_ch) = 0 Then Exit Do
        sym = sym + the_ch
        getch
    Loop
    the_num = Val(sym)
    symtype = tynum
End Sub

Sub readstr
    sym = Chr$(34)
    getch
    While the_ch <> Chr$(34)
        If the_ch = "" Then
            Print at_line$; "string not terminated": errors = true
            Exit Sub
        End If
        sym = sym + the_ch
        getch
    Wend
    getch
    symtype = tystring
End Sub

Sub skiptoeol
    textp = Len(thelin) + 1
    the_ch = ""
    sym = ""
    symtype = tyunknown
End Sub

Function gettoeol$
    Dim s As String
    s = ""
    While the_ch <> ""
        s = s + the_ch
        getch
    Wend
    Call getsym
    gettoeol$ = s
End Function

' symtype: unknown, tystring, tynum, tyident, tystrident
' sym: the symbol just read, above, and punctuation
Sub getsym
    'print "in getsym"
    'dim ttt as double
    'ttt = timer
    'nsyms = nsyms + 1
    sym = ""
    symtype = tyunknown
    ' skip white space
    While the_ch <= " "
        If the_ch = "" Then Exit Sub Else getch
    Wend
    sym = the_ch
    Select Case ctype_arr(Asc(the_ch))
        Case ct_punc1: getch 'punctuation
        Case ct_alpha: readident 'identifier
        Case ct_digit, ct_period: readnumber 'number
        Case ct_dquote: readstr 'double quote
        Case ct_squote: skiptoeol 'comment
        Case ct_lt
            getch
            If InStr("=><", the_ch) > 0 Then sym = sym + the_ch: getch '<=, <> <<
        Case ct_gt
            getch
            If the_ch = "=" Or the_ch = ">" Then sym = sym + the_ch: getch ' >=, >>
        Case ct_amp:
            getch
            If the_ch = "H" Or the_ch = "h" Then
                readhex
            Else
                Print at_line$; "getsym: '& found, expecting 'h' but found:"; the_ch: errors = true
            End If
        Case Else
            Print at_line$; "getsym: unexpected character read:"; the_ch: errors = true
            getch
    End Select
    'scantime = scantime + (timer - ttt)
End Sub

Sub initgetsym (n As Long, col As Integer)
    'print "initgetsym"
    curline = n
    textp = col
    thelin = pgm(curline)
    the_ch = " "
    Call getsym
End Sub


